;----------------------------------------------------------------------------
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;* Author: Sanjib Sharma                                                     *
;* Copyright (c) 2012 Sanjib Sharma                                          *
;* All rights reserved.                                                      *
;*                                                                           *
;* This file is part of EBF.  The full EBF copyright notice, including       *
;* terms governing use, modification, and redistribution, is contained in    *
;* the files COPYING and COPYRIGHT, which can be found at the root           *
;* of the source code distribution tree.                                     *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;-----------------------------------------------------------------------------
;+
; NAME:
;       EBF_READ
;
; PURPOSE:
;       This procedure reads data from an ebf-format files in the form of a
;       variable or an idl anonymous structure. 
;
; DESCRIPTION
;       Data objects in an ebf file are stored like unix filenames
;       starting with / as root directory e.g.  '/FeH',  '/SDSS/g' ,
;       '/SDSS/r'. Using this analogy files correspond to data objects. 
;       
;
; CALLING SEQUENCE:
;       EBF_READ, filename, label, data [, ALIASES=Aliases] [, /APPEND] [,  /REORDER] [, STATUS=Status]
;
; PARAMS:
;    filename- A string specifying the filename to read
;    label-    A string specifying the path (if ending with "/") or the name of
;              the data object to be read from the ebf file.
;
; RETURN VALUE:
;       data: An array if tagname does not end with '/' or else a
;       structure containg the data objects in specified path.
;       See also APPEND option.
;
;       STATUS :  In event of I/O failure this is 0 else 1
;
;
; KEYWORDS:
;       ALIASES : This makes it possible to rename the tags in the loaded
;                 structure for example naming 'feh' as 'metals'. The
;                 number of tagnames (labels) should be equal to number of
;                 aliases. When multiple labels are read this must be set.  
;       APPEND :  Will append items to an existing structure or hash table
;                 (experimental) insteading of creating a new variable. Structure
;                 should not be named as XXX
;
;       REORDER : Will transpose the loaded data if it is a
;                 multidimenisonal array of fundamental data
;                 types (int,float etc). Structure arrays or fields 
;                 of structures cannot be transposed. 
;
;       DATAUNIT: If units exist then they will be returned in this
;                 keyword or else a blank string.
;
; EXAMPLE:
;
;       write some data
;
;       IDL> x=findgen(10) & y=findgen(10) &  z=findgen(10)
;       IDL> ebf_write,'test.ebf','/data1/x1',x
;       IDL> ebf_write,'test.ebf','/data1/y1',y,/APPEND
;       IDL> ebf_write,'test.ebf','/data1/z1',z,/APPEND
;       IDL> ebf_write,'test.ebf','/data1/data2/x1',x*2,/APPEND
;       IDL> ebf_write,'test.ebf','/data1/data2/y1',y*2,/APPEND
;       IDL> ebf_write,'test.ebf','/data1/data2/z1',z*2,/APPEND
;
;       get a summary of the file
;
;       IDL> ebf_info,'test.ebf'
;       
;
;       Read a variable::
;
;       IDL> ebf_read,'test.ebf','/data1/z1',myvar
;       Data is loaded such that myvar is z
;
;       read multiple variables into a structure
;
;       IDL> ebf_read,'test.ebf','/data1/',mystruct
;
;       this will load '/data1/x1', '/data1/y1' and '/data1/z1' in
;       structure mystruct such that
;
;       mystruct.x1 is x
;       mystruct.y1 is y  
;       mystruct.z1 is z
;
;       The field names of the structures are obtained by stripping
;       the path from the data name. Data objects whose
;       names contain '.' are excluded. 
;
;       read specific items into a structure
;
;       IDL> ebf_read,'test.ebf',['/data1/x1','/data1/data2/z1','/FeH'],mystruct,ALIASES=['x','z]
;       This loads such that       
;       mystruct.x is x  and     
;       mystruct.z is z*2
;
;       write a structure and read it back
;
;       IDL> mystruct1={x:x,y:y,z:z}
;       IDL> ebf_write,'test.ebf','/mystruct1',mystruct1
;       IDL> ebf_read,'test.ebf','/mystruct1',mystruct11
;
;       write a nested structure and read it back
;
;       IDL> mystruct2={x:x,y:y,z:z,mystr1:mystruct1}
;       IDL> ebf_write,'test.ebf','/mystruct1',mystruct2
;       IDL> ebf_read,'test.ebf','/mystruct1',mystruct22
;
;       Split a structure into a homogeneous arrays and write and read
;
;       IDL> mystruct1={x:x,y:y,z:z}
;       IDL> ebf_write,'test.ebf','/foo/',mystruct1
;
;       written as '/foo/x , /foo/y and /foo/z'
;
;       IDL> ebf_read,'test.ebf','/foo/',mystruct2
;
;       Write and read specific tagnames only 
;
;       IDL> mystruct1={x:x,y:y,z:z,p:p}
;       IDL> ebf_write,'test.ebf','/foo/',mystruct1,tagnames=["x","z"]
;       IDL> ebf_read,'test.ebf','/foo/'+["x","z"],mystruct2
;
;
;       Write and read specific tagnames only but with aliases
;
;       IDL> mystruct1={x:x,y:y,z:z,p:p}
;       IDL> ebf_write,'test.ebf','/foo/',mystruct1,tagnames=["x","z"],aliases=['x1','y1']
;       IDL> ebf_read,'test.ebf','/foo/'+["x1","z1"],mystruct2,aliases=['x2','z2']
;
;
; MODIFICATION HISTORY:
;       Written, 2011, Sanjib Sharma
;-


function ebf_getchecksum,filename
  checksum=0LL
  openr,lun,filename,error=err,/get_lun
  if err ne 0 then begin
     print,'file not found:',filename
     if n_elements(lun) ne 0 then free_lun,lun
  endif else begin
     if ~EOF(lun) then begin
        ebfheader_read,ebfh,lun
        if (string(ebfh.name) eq '/.ebf/info') and (ebfh.datatype eq 3) then begin 
           readu,lun,checksum
           if ebfh.flagswap eq 1 then swap_endian_inplace,checksum
        endif
     endif
     free_lun,lun     
  endelse
  return,checksum
end


;-------------------------------------------------------
;-------------------------------------------------------

;+
; NAME:
;       EBF_READSINGLE
;
; PURPOSE:
;       This function reads a single data object from and ebf-format files 
;
; DESCRIPTION
;       Data objects in an ebf file are stored like unix filenames
;       starting with / as root directory e.g.  '/FeH',  '/SDSS/g' ,
;       '/SDSS/r'. Using this analogy files correspond to data objects. 
;       
;
; CALLING SEQUENCE:
;       data=ebf_readsingle,filename,label,data[,REORDER=,STATUS=]
; INPUTS:
;    filename- A string specifying the filename to read
;    label-    A string specifying the path (if ending with "/") or the name of
;              the data object to be read from the ebf file.
;
; OUTPUTS:
;       data: An array, structure or a scalar 
;
;       STATUS :  In event of I/O failure this is 0 else 1
;
;
; KEYWORDS:
;       REORDER : Will transpose the loaded data if it is a
;                 multidimenisonal array of fundamental data
;                 types (int,float etc). Structure arrays or fields 
;                 of structures cannot be transposed. 
;
;
; MODIFICATION HISTORY:
;       Written, 2011, Sanjib Sharma
;-


;------------------------------------------------------------------

;+
; NAME:
;       EBF_DESCR2STRUCT
;
; PURPOSE:
;       This function converts a string representation of a structure 
;       into an anonymous idl structure. Nested structures are allowed.
;       Note, structures should not contain pointers or strings.
;
; DESCRIPTION:
;       This is a recursive function to enable reading nested structures.      
;
; CALLING SEQUENCE:
;       mystruct=ebf_descr2struct(mystring)
; INPUTS:
;    s    -    A string specifying the structure
;
; OUTPUTS:
;       data: a structure
;
; KEYWORDS:
;       NAME : In case of nested structures it will read the structure 
;              specified by the particular name
;
; EXAMPLE:
;
; MODIFICATION HISTORY:
;       Written, 2011, Sanjib Sharma
;-

;; function ebf_sdef2struct,sdef
;; sdef1=strsplit(strtrim(sdef),string(10b),/extract)
;; words=strsplit(sdef1[0],' ',/extract)
;; if words[0] eq 'ver-1' then begin
;;    return ebf_descr2struct(strsplit(strtrim(sdef),' '+string(10b),/extract))
;; endif
;; if words[0] eq 'ver-2' then begin

;; i=1
;; temp=''
;; while temp ne '<sdef>' do begin
;; words=strsplit(sdef1[i],' ',/extract)
;; temp=words[0]   
;; i=i+1
;; endwhile
;; words=strsplit(sdef1[i],' ',/extract)
;; rank=long64(words[2])
;; nsize=long64(words[3+rank])
;; i=i+1
;; return ebf_sdef2struct_v2(sdef1,i,nsize)
;; endif
;; end

;; function ebf_sdef2struct_v2,sdef,start,nsize
;; i=0
;; while i lt nsize do begin
;; words=strsplit(sdef[start],' ',/extract)
;; name=words[0]
;; dtype=long64(words[1])
;; rank=long64(words[2])
;; dims=lon64arr(rank)
;; for j=0,rank-1 do dims[j]=long64(words[3+j])
;; nsize=long64(words[3+rank])
;; start=start+1
;;         case dtype of
;;            1: data=bytarr(product(dims,/integer))
;;            2: data=lonarr(product(dims,/integer))
;;            3: data=lon64arr(product(dims,/integer))
;;            4: data=fltarr(product(dims,/integer))
;;            5: data=dblarr(product(dims,/integer))
;;            6: data=intarr(product(dims,/integer))
;;            9: data=bytarr(product(dims,/integer))
;;            10: data=bytarr(product(dims,/integer))
;;            11: data=uintarr(product(dims,/integer))
;;            12: data=ulonarr(product(dims,/integer))
;;            13: data=ulon64arr(product(dims,/integer))
;;            8:begin 
;;               data=replicate(ebf_descr2struct(sdef,start),product(dims))
;;            end
;;            else: begin 
;;               message,'data type not recognized/supported:'+dtype
;;            end
;;         endcase
;;         if dtype eq 9 then print,'EBF WARNING: reading int8 as uint8'
;;         if n_elements(dims) gt 1 then begin
;;            dims=reverse(dims)
;;            data=reform(data,dims,/overwrite)
;;            data=reform(data,/overwrite) ; remove extra dims
;;         endif else begin
;;            if dims[0] eq 1 then data=data[0]
;;         endelse
;;         if n_tags(a) eq 0 then a=create_struct(dname,data) else a=create_struct(a,dname,data)
;; i=i+1
;; endwhile
;; return, a
;; end


function ebf_line2ftag,line
on_error,2
words=strtrim(strsplit(line,',',/extract),2)
;words=strsplit(line,' ',/extract)
rank=long64(words[2])
dims=lon64arr(rank>1)
for j=0,rank-1 do dims[j]=long64(words[3+j])
if rank eq 0 then dims[0]=1
if n_elements(words) gt (4+rank) then units=words[4+rank] else units=''
tag={name:words[0],etype:long64(words[1]),rank:rank,dims:reverse(dims),nsize:long64(words[3+rank]),unit:units}
return, tag
end

function ebf_sdef2struct_v2,sdef,start,nsize,dataunit=dataunit
on_error,2
status=0
if start eq 0 then begin
status=1
temp=''
while temp ne '<sdef>' do begin
;words=strsplit(sdef[start],' ',/extract)
words=strtrim(strsplit(sdef[start],',',/extract),2)
temp=words[0]   
start=start+1
endwhile
ftag=ebf_line2ftag(sdef[start])
start=start+1
nsize=ftag.nsize
endif
dataunit=0

i=0
while i lt nsize do begin
ftag=ebf_line2ftag(sdef[start])
start=start+1
unit1=ftag.unit
        case ftag.etype of
           1: data=bytarr(product(ftag.dims,/integer))
           2: data=lonarr(product(ftag.dims,/integer))
           3: data=lon64arr(product(ftag.dims,/integer))
           4: data=fltarr(product(ftag.dims,/integer))
           5: data=dblarr(product(ftag.dims,/integer))
           6: data=intarr(product(ftag.dims,/integer))
           9: data=bytarr(product(ftag.dims,/integer))
           10: data=bytarr(product(ftag.dims,/integer))
           11: data=uintarr(product(ftag.dims,/integer))
           12: data=ulonarr(product(ftag.dims,/integer))
           13: data=ulon64arr(product(ftag.dims,/integer))
           8:begin 
              data=replicate(ebf_sdef2struct_v2(sdef,start,ftag.nsize,dataunit=unit1),product(ftag.dims))
           end
           else: begin 
              message,'data type not recognized/supported:'+dtype
           end
        endcase
        if ftag.etype eq 9 then print,'EBF WARNING: reading int8 as uint8'
        if n_elements(ftag.dims) gt 1 then begin
           data=reform(data,ftag.dims,/overwrite)
           data=reform(data,/overwrite) ; remove extra dims
        endif else begin
           if ftag.dims[0] eq 1 then data=data[0]
        endelse
        if n_tags(a) eq 0 then a=create_struct(ftag.name,data) else a=create_struct(a,ftag.name,data)
        if n_tags(dataunit) eq 0 then dataunit=create_struct(ftag.name,unit1) else dataunit=create_struct(dataunit,ftag.name,unit1)
i=i+1
endwhile

if status eq 1 then begin
;words=strsplit(sdef[start],' ',/extract)
words=strtrim(strsplit(sdef[start],',',/extract),2)
if words[0] ne '</sdef>' then message,'Problem in structure definition'
endif
return, a
end

function ebf_sdef2struct_v1,svv
on_error,2
  i=0
;  print,svv
  while svv[i] ne '{' do i++ 
  i++
  while i lt n_elements(svv) do begin
     if svv[i] eq '}' then break
     k=i
     dtype=svv[i]
     if dtype eq 'struct' then begin
        i++
        l=1
        while l ne 0 do begin
           i++
           if svv[i] eq '}' then l-- 
           if svv[i] eq '{' then l++ 
        endwhile   
     endif
     dname=svv[i+1]
     rank=long64(svv[i+2])
     if rank gt 0 then begin
        dims=lon64arr(rank)
        ;; print, rank
        ;; print, svv[i+3]
        for j=0,rank-1 do dims[j]=long64(svv[i+j+3])
     endif else begin
        dims=1LL
     endelse
        case dtype of
           'char': data=bytarr(product(dims,/integer))
           'int32': data=lonarr(product(dims,/integer))
           'int64': data=lon64arr(product(dims,/integer))
           'float32': data=fltarr(product(dims,/integer))
           'float64': data=dblarr(product(dims,/integer))
           'int16': data=intarr(product(dims,/integer))
           'int8': data=bytarr(product(dims,/integer))
           'uint8': data=bytarr(product(dims,/integer))
           'uint16': data=uintarr(product(dims,/integer))
           'uint32': data=ulonarr(product(dims,/integer))
           'uint64': data=ulon64arr(product(dims,/integer))
           'struct':begin 
              data=replicate(ebf_sdef2struct_v1(svv[k:*]),product(dims))
           end
           else: begin 
              message,'data type not recognized/supported:'+dtype
           end
        endcase
        if dtype eq 'int8' then print,'EBF WARNING: reading int8 as uint8'
        if n_elements(dims) gt 1 then begin
           dims=reverse(dims)
           data=reform(data,dims,/overwrite)
           data=reform(data,/overwrite) ; remove extra dims
        endif else begin
           if dims[0] eq 1 then data=data[0]
        endelse
        if n_tags(a) eq 0 then a=create_struct(dname,data) else a=create_struct(a,dname,data)

     if svv[i+2+rank+1] eq ';' then i=i+4+rank else message,'EBF error: missing ; in sdef'

  endwhile   

  return, a
end


function ebf_sdef2struct,sdef,dataunit=dataunit
on_error,2
sdef1=strsplit(strtrim(sdef),string(10b),/extract)
words=strsplit(sdef1[0],' ',/extract)
if words[0] eq 'ver-1' then begin
   return, ebf_sdef2struct_v1(strsplit(strtrim(sdef),' '+string(10b),/extract))
endif
if words[0] eq 'ver-2' then begin
return, ebf_sdef2struct_v2(sdef1,0,0,dataunit=dataunit)
endif
end

;; function ebf_descr2struct,svv
;;   i=0
;; ;  print,svv
;;   while svv[i] ne '{' do i++ 
;;   i++
;;   while i lt n_elements(svv) do begin
;;      if svv[i] eq '}' then break
;;      k=i
;;      dtype=svv[i]
;;      if dtype eq 'struct' then begin
;;         i++
;;         l=1
;;         while l ne 0 do begin
;;            i++
;;            if svv[i] eq '}' then l-- 
;;            if svv[i] eq '{' then l++ 
;;         endwhile   
;;      endif
;;      dname=svv[i+1]
;;      rank=long64(svv[i+2])
;;      if rank gt 0 then begin
;;         dims=lon64arr(rank)
;;         ;; print, rank
;;         ;; print, svv[i+3]
;;         for j=0,rank-1 do dims[j]=long64(svv[i+j+3])
;;      endif else begin
;;         dims=1LL
;;      endelse
;;         case dtype of
;;            'char': data=bytarr(product(dims,/integer))
;;            'int32': data=lonarr(product(dims,/integer))
;;            'int64': data=lon64arr(product(dims,/integer))
;;            'float32': data=fltarr(product(dims,/integer))
;;            'float64': data=dblarr(product(dims,/integer))
;;            'int16': data=intarr(product(dims,/integer))
;;            'int8': data=bytarr(product(dims,/integer))
;;            'uint8': data=bytarr(product(dims,/integer))
;;            'uint16': data=uintarr(product(dims,/integer))
;;            'uint32': data=ulonarr(product(dims,/integer))
;;            'uint64': data=ulon64arr(product(dims,/integer))
;;            'struct':begin 
;;               data=replicate(ebf_descr2struct(svv[k:*]),product(dims))
;;            end
;;            else: begin 
;;               message,'data type not recognized/supported:'+dtype
;;            end
;;         endcase
;;         if dtype eq 'int8' then print,'EBF WARNING: reading int8 as uint8'
;;         if n_elements(dims) gt 1 then begin
;;            dims=reverse(dims)
;;            data=reform(data,dims,/overwrite)
;;            data=reform(data,/overwrite) ; remove extra dims
;;         endif else begin
;;            if dims[0] eq 1 then data=data[0]
;;         endelse
;;         if n_tags(a) eq 0 then a=create_struct(dname,data) else a=create_struct(a,dname,data)

;;      if svv[i+2+rank+1] eq ';' then i=i+4+rank else message,'EBF error: missing ; in sdef'

;;   endwhile   

;;   return, a
;; end



function ebf_readsingle,fname1,name,REORDER=Reorder,STATUS=Status ,VERBOSE=Verbose,DATAUNIT=dataunit,LOCATION=location,nochar=nochar
ON_ERROR,2
;  Verbose=1
;  if keyword_set(Location) then loc=location else
;  loc=ebfmap_getlocation(fname1,name)
  if keyword_set(Location) then loc=location else loc=ebftable_get(fname1,name)
  status=0
  openr,lun,fname1,error=err,/get_lun ;,bufsize=1024l*1024l  
  if err ne 0 then begin
     print,'file not found:',fname1
     if n_elements(lun) ne 0 then free_lun,lun
  endif else begin

  if loc lt 0 then begin 
     free_lun,lun
     message,'data object:'+name+' could not be located in '+fname1
  endif   

;-------------------------------------------------------

     point_lun,lun,loc
                                ; read the header
     ebfheader_read,ebfh,lun
     dataunit=ebfh.dataunit
     if(ebfh.rank lt 0) then begin 
        free_lun,lun
        message,'Rank le 0, corrupted ebf file'
     endif   

;     elements=product(ebfh.dim[0:ebfh.rank-1],/integer)
     elements=product(ebfh.dim,/integer)
     if(strlowcase(string(ebfh.name)) ne name) then begin
        print, strlowcase(string(ebfh.name)),name
        message,'Cannot find data'
     endif
        case ebfh.datatype of
           1: data=bytarr(elements)
           2: data=lonarr(elements)
           3: data=lon64arr(elements)
           4: data=fltarr(elements)
           5: data=dblarr(elements)
           6: data=intarr(elements)
           8: begin 
;              ebf_read,fname1,name+'.sdef',descr              
              ;; ind=where(ebfh.keyword eq 'sdef',count)
              ;; if count ne 1 then message,'Ebf error, structure defintion not found'
              ;; descr=ebfh.value[ind[0]]
              if (strmid(ebfh.sdef,0,5) ne 'ver-1')and(strmid(ebfh.sdef,0,5) ne 'ver-2') then begin
                  free_lun,lun     
                  message,'Ebf error: sdef version should be 1'    
              endif   
;              data=replicate(ebf_descr2struct(strsplit(ebfh.sdef,' '+string(10b),/extract)),elements)
              data=replicate(ebf_sdef2struct(ebfh.sdef,dataunit=dataunit),elements)
           end
; comment or uncomment for  int8 support
            9: data=bytarr(elements)
           10: data=bytarr(elements)
           11: data=uintarr(elements)
           12: data=ulonarr(elements)
           13: data=ulon64arr(elements)
           else:message,'data type not recognized :'+string(ebfh.datatype)
        endcase
        if ebfh.datatype eq 9 then print,'EBF WARNING: reading int8 as uint8'
           
                                ; read and reform data
        readu,lun,data
        if ebfh.flagswap eq 1 then swap_endian_inplace,data
        if(ebfh.rank gt 1) then begin 
           data=reform(data,reverse(ebfh.dim),/overwrite)
           if keyword_set(REORDER) then begin
              if ebfh.datatype eq 8 then begin 
                 free_lun,lun
                 message,'A structure array cannot be  transposed'
              endif   
              data=transpose(data) ; reverse array dims
           endif
        endif else begin
;           if (ebfh.rank eq 1)and(ebfh.dim[0] eq 1) then data=data[0] 
           if (ebfh.rank le 1)and(elements eq 1) then data=data[0] 
        endelse   
        status=1                    
        if keyword_set(Verbose) then print,'loaded ->',fname1,':',name

     free_lun,lun     
     if status eq 0 then begin
        print,name,' not found'
     endif
  endelse
;  if ebfh.datatype eq 1 then return,string(data) else return,data
  if (ebfh.datatype eq 1)and(~keyword_set(nochar)) then data=string(data) 
  return,data
end


;-------------------------------------------------------
;-------------------------------------------------------

pro ebf_splitadd,mydata,data,tname,labels
  if (size(data,/n_dim) ne 2) then message,'Data must be of rank 2 to split'
  if ((size(data,/dim))[0] eq 3)and((size(data,/dim))[1] eq 3) then message,'Data is 3x3 and is ambiguous'
  if ((size(data,/dim))[0] ne 3)and((size(data,/dim))[1] ne 3) then message,'one of the dimension should be of size 3'
  if ((size(data,/dim))[0] eq 3) then data=transpose(data)


  if (strupcase(tname) eq '/POS3') then begin 
     if size(mydata,/sname) eq 'XXX' then mydata = CREATE_STRUCT('px', data[*,0]) else  mydata = CREATE_STRUCT(mydata, 'px', data[*,0])
     mydata = CREATE_STRUCT(mydata, 'py', data[*,1])
     mydata = CREATE_STRUCT(mydata, 'pz', data[*,2])     
     labels=['px','py','pz']
  endif
  if strupcase(tname) eq '/VEL3' then begin 
     if size(mydata,/sname) eq 'XXX' then mydata = CREATE_STRUCT('vx', data[*,0]) else     mydata = CREATE_STRUCT(mydata, 'vx', data[*,0])
     mydata = CREATE_STRUCT(mydata, 'vy', data[*,1])
     mydata = CREATE_STRUCT(mydata, 'vz', data[*,2])
     labels=['vx','vy','vz']
  endif
end





pro ebf_read,fname1,path,mydata,APPEND=Append,DATAUNIT=dataunit,REORDER=reorder,Aliases=Aliases,Split=Split,STATUS=Status,VERBOSE=Verbose,VERSION=Version,nochar=nochar
;ON_ERROR,2
  status=0
  Version='0.0.3'

  path1=strtrim(strlowcase(path),2)

  if (n_elements(path1) ne n_elements(aliases))and(keyword_set(aliases)) then message,'number of Tagnames not equal to number of aliases'
; if append and given varibale not structure then delete it
;     if keyword_set(APPEND)and(size(mydata,/type) ne 8) then mydata={XXX,no:0}

;        if (size(mydata,/type) ne 8)or(~keyword_set(APPEND)) then mydata={XXX,no:0}


  if n_elements(path1) gt 1 then begin
                                ; mutiple items    
;     loc=ebfmap_getlocation(fname1,path1)
;     loc=ebftable_get(fname1,path1)
     loc=lon64arr(n_elements(path1))
     for i=0,n_elements(ind)-1 do begin
        loc[i]=ebftable_get(fname1,keys[path1[i]])
     endfor   

     if(n_elements(path1) eq total(loc ge 0,/integer)) then begin
        unused=0
        if ~keyword_set(aliases) then begin 
           aliases=path1
           for i=0,n_elements(path1)-1 do begin
              aliases[i]=strmid(path1[i],strpos(path1[i],'/',/reverse_search)+1)
           endfor   
        endif   
     endif else message,'Some tags could not be found'
  endif else begin
                                ; mutiple items as path ending with '/'   
     temp=byte(path1[0]) 
     if (temp[n_elements(temp)-1] eq 47) then begin 
        path1=path1+'*'
        temp=byte(path1[0]) 
     endif
     if (temp[n_elements(temp)-1] eq 42) then begin
        if n_elements(path1) gt 1 then message,"Only one tag ending with * allowed"
        path1=path1[0]
        path2=strmid(path,0,strpos(path1,'/',/reverse_search)+1)
;        keys=ebfmap_keys(fname1)
        keys=ebf_getkeys(fname1)
        ind=where((strmatch(keys,path1,/FOLD_CASE) eq 1)and(strmatch(keys,path2+"*/*",/FOLD_CASE) eq 0)and(strmatch(keys,path2+"*.*",/FOLD_CASE) eq 0)and(strmatch(keys,path2+".*",/FOLD_CASE) eq 0),count)
        if count gt 0 then begin 
           path1=keys[ind]
;           loc=ebfmap_getlocation(fname1,keys[ind])
           loc=lon64arr(n_elements(ind))
           for i=0,n_elements(ind)-1 do begin
              loc[i]=ebftable_get(fname1,keys[ind[i]])
           endfor   
           aliases=strlowcase(strmid(keys[ind],strlen(path2)))

           ind=sort(loc)
           path1=path1[ind]
           aliases=aliases[ind]
           loc=loc[ind]
        endif else begin
           print,fname1,path1
           print,keys
           message,'no matching data found'
        endelse           
     endif else begin
                                ; a single item   
;        loc=ebfmap_getlocation(fname1,path1)
        loc=ebftable_get(fname1,path1)
        if loc lt 0 then begin 
           print,"Tag name not found: ",path1
        endif else begin  
           path2=strmid(path,0,strpos(path1,'/',/reverse_search)+1)
           if(~keyword_set(aliases)) then aliases=strmid(path1,strlen(path2))
        endelse   
     endelse   
  endelse  

;  if items> 1 or append only then add to structure or else create a
;  simple variable
  if (n_elements(path1) gt 1)or(keyword_set(APPEND))  then begin
     dataunit2={UUU,no:0}
;     dataunit2=""
                                ; if appending to hash they must already exist or else make a struct
     if (size(mydata,/type) ne 11) then begin
        if (size(mydata,/type) ne 8)or(~keyword_set(APPEND)) then begin 
           mydata={XXX,no:0}
        endif
        for i=0,n_elements(path1)-1 do begin
                                ; split POS3 VEL3
           if keyword_set(SPLIT)and((strupcase(path1[i]) eq '/POS3')or(strupcase(path1[i]) eq '/VEL3'))  then begin 
              data=ebf_readsingle(fname1,path1[i],REORDER=reorder,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit1,nochar=nochar);location=loc[i]
              ebf_splitadd,mydata,data,path1[i],labels
;              dataunit2=[dataunit2,dataunit1,dataunit1,dataunit1]
              for j=0,n_elements(labels)-1 do begin
                 if size(dataunit2,/sname) eq 'UUU' then begin
                    dataunit2=create_struct(labels[j],dataunit1)     
                 endif else begin
                    dataunit2=create_struct(dataunit2,labels[j],dataunit1)     
                 endelse
              endfor
           endif else begin
              if size(mydata,/sname) eq 'XXX' then begin
                 mydata = CREATE_STRUCT(aliases[i],ebf_readsingle(fname1,path1[i],REORDER=reorder,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit1,nochar=nochar)) 
              endif else begin                
                 mydata = CREATE_STRUCT(mydata,aliases[i],ebf_readsingle(fname1,path1[i],REORDER=reorder,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit1,nochar=nochar))     
              endelse
;              dataunit2=[dataunit2,dataunit1]     
              if size(dataunit2,/sname) eq 'UUU' then begin
                 dataunit2=create_struct(aliases[i],dataunit1)     
              endif else begin
                 dataunit2=create_struct(dataunit2,aliases[i],dataunit1)     
              endelse

           endelse
        endfor        
     endif else begin  

        if obj_isa(mydata,'hashtable') then begin
           if (~keyword_set(APPEND))and(mydata->count() gt 0) then begin
              obj_destroy,mydata
              mydata=obj_new('hashtable',/no_duplicates)
              print,'Ebf Warning:Preferably pass an empty hashtable.' 
              print,'As ebf_read was not called in append mode, obj_destroy is beign used.'
              print,'If any of the contained objects are heap data or objects,'
              print,'the user is responsible for freeing those pointers or objects. For IDL VER ge 8.0 this should not be a problem due to garbage collection'
;              message,'Not in append mode, hence empty the hashtable first'
           endif   
           for i=0,n_elements(path1)-1 do begin
              mydata->add,aliases[i],ebf_readsingle(fname1,path1[i],REORDER=reorder,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit1,nochar=nochar)     
;              dataunit2=[dataunit2,dataunit1]
              if size(dataunit2,/sname) eq 'UUU' then begin
                 dataunit2=create_struct(aliases[i],dataunit1)     
              endif else begin
                 dataunit2=create_struct(dataunit2,aliases[i],dataunit1)     
              endelse
           endfor
           
        endif   

        if obj_isa(mydata,'hash') then begin
           if (~keyword_set(APPEND))and(n_elements(mydata) gt 0) then begin
              mydata->remove,/ALL
           endif   
           for i=0,n_elements(path1)-1 do begin
              mydata[aliases[i]]=ebf_readsingle(fname1,path1[i],REORDER=reorder,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit1,nochar=nochar)
;              dataunit2=[dataunit2,dataunit1]
              if size(dataunit2,/sname) eq 'UUU' then begin
                 dataunit2=create_struct(aliases[i],dataunit1)     
              endif else begin
                 dataunit2=create_struct(dataunit2,aliases[i],dataunit1)     
              endelse
           endfor        
        endif   
     endelse
     
;     if n_elements(dataunit2) gt 1   then   dataunit=dataunit2[1:*]   
     dataunit=dataunit2

  endif else begin
     mydata=ebf_readsingle(fname1,path1,REORDER=reorder,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit,nochar=nochar)
  endelse

end




