;----------------------------------------------------------------------------
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;* Author: Sanjib Sharma                                                     *
;* Copyright (c) 2012 Sanjib Sharma                                          *
;* All rights reserved.                                                      *
;*                                                                           *
;* This file is part of EBF.  The full EBF copyright notice, including       *
;* terms governing use, modification, and redistribution, is contained in    *
;* the files COPYING and COPYRIGHT, which can be found at the root           *
;* of the source code distribution tree.                                     *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;-----------------------------------------------------------------------------
;+
;
; NAME:
;       EBF_RENAME
;
; PURPOSE: 
;       Rename or remove a data item in an ebf file.        
;
; CALLING SEQUENCE:
;       ebf_rename,filename,old_key, new_key
;
; PARAMS:
;       filename : A string specifying the filename to write to.
;       old_key  : A string, name of key to rename.
;       new_key  : A string, the new data name. If new_key 
;                  is an empty string then the data item is removed. 
;                  Remove simply means putting the data in the trash 
;                  folder /.tr/ of the ebf file using rename.
;                  For this a name of the form '/.tr'+oldkey+'.X' is
;                  created. Here X is a an integer greater than equal
;                  to zero, which is incremented each time the item
;                  with same name is deleted. 
;
; EXAMPLE:
;       ebf_rename,'check.ebf','/x1','/x2'
;       ebf_rename,'check.ebf','/x1',''    
;
; EXTERNAL PROGRAMS USED:
;       ebftable_remove.pro, ebftable_get.pro, ebftable_put.pro
;
; MODIFICATION HISTORY:
;       Written, 2013, Sanjib Sharma
;-
    

pro ebf_rename,filename,oldkey1,newkey1
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  oldkey=strtrim(strlowcase(oldkey1),2)
  newkey=strtrim(strlowcase(newkey1),2)
  loc=ebftable_get(filename,oldkey)                               
  if (newkey ne oldkey)and(loc ge 0) then begin            
     if newkey eq '' then begin
        i=0LL
        while ebftable_get(filename,'/.tr'+oldkey+'.'+strcompress(string(i),/remove_all)) ne -1 do begin
           i=i+1LL
           if i gt 1000000LL then message,'EBF: error, too many deleted items'
        endwhile
        newkey='/.tr'+oldkey+'.'+strcompress(string(i),/remove_all))
     endif

     loc1=ebftable_get(filename,newkey)                               
     if loc lt 0 then message,'EBF error: data item/key not found'        
     if loc1 gt 0 then message,'EBF error: a key with given name already exists'
     
     ;; rename 
     openu,lun,filename,/get_lun,err=err
     if err ne 0 then begin
        ecode=1000
        message,'file not found: '+filename
     endif else begin
        point_lun,lun,loc
        ebfheader_read,ebfh,lun
        ebfheader_rename,ebfh
        point_lun,lun,loc
        ebfheader_write,ebfh,lun
     endif
     if n_elements(lun) ne 0 then free_lun,lun

     ebftable_remove(filename,oldkey)
     ebftable_put(filename,newkey,loc)
  endif
end
