;----------------------------------------------------------------------------
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;* Author: Sanjib Sharma                                                     *
;* Copyright (c) 2012 Sanjib Sharma                                          *
;* All rights reserved.                                                      *
;*                                                                           *
;* This file is part of EBF.  The full EBF copyright notice, including       *
;* terms governing use, modification, and redistribution, is contained in    *
;* the files COPYING and COPYRIGHT, which can be found at the root           *
;* of the source code distribution tree.                                     *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;-----------------------------------------------------------------------------
;+
;
; NAME:
;       EBF_WRITE
;
; PURPOSE: 
;          Writes a data specified by a tag name to a .ebf files.   
;
; DESCRIPTION:
;       Data objects in an ebf file are stored like unix filenames
;       starting with / as root directory e.g.  '/FeH',  '/SDSS/g' ,
;       '/SDSS/r'. Using this analogy files correspond to data objects. 
;       
; CALLING SEQUENCE:
;       ebf_write,filename,label,data,[TAGNAMES=,ALIASES=,APPEND=,REORDER=,STATUS=,DATAUNIT=]
;
; EXAMPLES:
;         see examples in ebf_read.pro
;
; KEYWORDS:
;
;       tagname: If one only wants to write specific items 
;                 then these can be specified as string array
;
;       aliases: This makes it possible to assign labels in the
;                 ebf file different from structure filed names
;                 structure for example naming 'mystruct.feh' 
;                  as '"/metals'. The number of aliases should be
;                  equal to number of items being written.
;
;       append :  To append an item to an existing file
;
;       reorder : To transpose and write a multidimensional array. A
;                 structure array or its elements cannot be transposed.
;
;       dataunit: a string or an array of string specifying units.
;                 The number of units should be equal to number of
;                 items being written
;
;       status :  In event of I/O failure this is 0 else 1
;
; PARAMS:
;
;       filename : A string specifying the filename to write to.
;
;       tagname :  A string specifying the tagname of the data to be
;       written
;
;       data :     Data to be stored. Single or mutidimensional arrays of long,
;                 long64, float, double and byte types. Or
;                 structures. Recursive structures are also allowed.
;
; EXTERNAL PROGRAMS USED:
;       ebf_struct2descr.pro:
;
; MODIFICATION HISTORY:
;       Written, 2010, Sanjib Sharma
;-




pro ebf_write_single,foutput,name,data,REORDER=Reorder,STATUS=Status,VERBOSE=Verbose,APPEND=Append,DATAUNIT=dataunit,nochar=nochar
on_error,2

  status=0
  datasize=[0,1,4,8,4,8,2,2,0,1,1,2,4,8]
  if size(data,/type) eq 7 then begin
     data1=byte(data)
     ebfheader_set,ebfh,name,data1,REORDER=Reorder,DATAUNIT=dataunit,DTYPE=1
  endif else begin
     ebfheader_set,ebfh,name,data,REORDER=Reorder,DATAUNIT=dataunit,nochar=nochar
  endelse


  if((ebfh.datatype ne 0)and(ebfh.rank ge 0)and(ebfh.rank le 8)) then begin
        if keyword_set(Append) then begin
           if ebftable_get(foutput,name) ge 0 then begin
              message,"Ebf error: data object already present"+name
           endif   
        endif


     openw,lun,foutput,error=err,APPEND=Append,/GET_LUN;,bufsize=1024l*1024l
     if(err ne 0) then begin
        print,err
        print,!ERROR_STATE.msg
        if n_elements(lun) ne 0 then free_lun,lun
     endif else begin         
        status=0
        point_lun,-lun,pos

        if keyword_set(Append) then begin
        endif

;        ebf_write_header,ebfh,lun
        ebfheader_write,ebfh,lun
        if size(data,/type) eq 7 then begin
           if keyword_set(REORDER) then writeu,lun,transpose(data1) else writeu,lun,data1
        endif else begin   
           if keyword_set(REORDER) then writeu,lun,transpose(data) else writeu,lun,data
        endelse   
        status=1
        free_lun,lun

        if keyword_set(Append) then begin
           ebftable_put,foutput,name,pos
        endif   

     endelse
     if keyword_set(Verbose) then print,'written-> ',foutput,':',string(name)
  endif else begin
     print,'data not compatible to be written'
     print,'rank=',ebfh.rank
     print,'datatype=',ebfh.datatype
  endelse


end

function ebf_getcheck_units,tagnames,dataunit=dataunit
dataunit1=''
if n_elements(tagnames) gt 0 then begin
  if keyword_set(Dataunit) then begin 
     dataunit1=dataunit 
     if size(dataunit1,/type) eq 7 then begin
        dataunit1=create_struct(tagnames[0],dataunit[0])
        for i=1,n_elements(tagnames)-1 do begin
           dataunit1=create_struct(dataunit1,tagnames[i],dataunit[i])
        endfor        
     endif else begin
        if n_tags(dataunit1) ne n_elements(tagnames) then message,'number of Tagnames not equal to number of dataunit'
     endelse
  endif else begin
     dataunit1=create_struct(tagnames[0],'')
     for i=1,n_elements(tagnames)-1 do begin
        dataunit1=create_struct(dataunit1,tagnames[i],'')
     endfor
  endelse
endif 
return,dataunit1
end


pro ebf_write_multi,fname1,path,mydata,Tagnames=Tagnames,Aliases=Aliases,REORDER=reorder,APPEND=Append,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit,nochar=nochar
on_error,2
  status=0
  if (size(mydata,/type) ne 8) and (size(mydata,/type) ne 11) then message,'data must be a structure or hash map'
  temp=byte(path) 
  if temp[n_elements(temp)-1] ne 47 then message,'keyword path must end with /'+path


  if size(mydata,/type) eq 8 then  begin
;     if size(mydata,/n_elements) ne 1 then message,'data must not be an array'
     datatags=strlowcase(tag_names(mydata))
     if ~keyword_set(tagnames) then tagnames=datatags
     if keyword_set(aliases) then aliases1=aliases else aliases1=tagnames
     aliases1=strlowcase(aliases1)
     if (n_elements(tagnames) ne n_elements(aliases1)) then message,'number of Tagnames not equal to number of aliases'
     dataunit1=ebf_getcheck_units(tagnames,dataunit=dataunit)
;     if keyword_set(Dataunit) then dataunit1=dataunit else dataunit1=strarr(n_elements(tagnames))
;     if (n_elements(tagnames) ne n_tags(dataunit1)) then message,'number of Tagnames not equal to number of dataunit'
     if keyword_set(Append) eq 0 then append=0

     for i=0,n_elements(tagnames)-1 do begin
        ind=where(strcmp(datatags,tagnames[i],/FOLD_CASE),count)
        if count eq 1 then begin
           name=byte(aliases1[i]) 
           if (size(mydata.(ind[0]),/type) eq 8)or (size(mydata.(ind[0]),/type) eq 11) then begin
              ebf_write_multi,fname1,path+aliases1[i]+'/',mydata.(ind[0]),REORDER=reorder,APPEND=Append,DATAUNIT=dataunit1.(i),nochar=nochar
           endif else begin
              ebf_write_single,fname1,path+string(name),mydata.(ind[0]),APPEND=append,REORDER=Reorder,DATAUNIT=dataunit1.(i),nochar=nochar
           endelse
           if append eq 0 then append=1
           status=1
        endif 
        if count gt 1 then message,'EBF error: multiple tags'
     endfor

  endif else if obj_isa(mydata,'hashtable') then begin 
           datatags=(mydata->keys())
           if ~keyword_set(tagnames) then tagnames=datatags
           if keyword_set(aliases) then aliases1=aliases else aliases1=tagnames
           aliases1=strlowcase(aliases1)
           if (n_elements(tagnames) ne n_elements(aliases1)) then message,'number of Tagnames not equal to number of aliases'
           dataunit1=ebf_getcheck_units(tagnames,dataunit=dataunit)
;           if keyword_set(Dataunit) then dataunit1=dataunit else dataunit1=strarr(n_elements(tagnames))
;           if (n_elements(tagnames) ne n_tags(dataunit1)) then message,'number of Tagnames not equal to number of dataunit'
           if keyword_set(Append) eq 0 then append=0

           for i=0,n_elements(tagnames)-1 do begin
              ind=where(strcmp(datatags,tagnames[i],/FOLD_CASE),count)
              if count eq 1 then begin
                 name=byte(aliases1[i]) 
                 if (size(mydata->get(datatags[ind[0]]),/type) eq 8)or (size(mydata->get(datatags[ind[0]]),/type) eq 11) then begin
                    ebf_write_multi,fname1,path+aliases1[i]+'/',mydata->get(datatags[ind[0]]),REORDER=reorder,APPEND=Append,DATAUNIT=dataunit1.(i),nochar=nochar
                 endif else begin                 
                    ebf_write_single,fname1,path+string(name),mydata->get(datatags[ind[0]]),APPEND=append,REORDER=Reorder,DATAUNIT=dataunit1.(i),nochar=nochar
                 endelse
                 if append eq 0 then append=1
                 status=1
              endif 
              if count gt 1 then message,'EBF error: multiple tags'
           endfor

    endif  else if obj_isa(mydata,'hash') then begin 
           datatags=((mydata->keys())->toArray())
           if ~keyword_set(tagnames) then tagnames=datatags
           if keyword_set(aliases) then aliases1=aliases else aliases1=tagnames
           aliases1=strlowcase(aliases1)
           if (n_elements(tagnames) ne n_elements(aliases1)) then message,'number of Tagnames not equal to number of aliases'
           dataunit1=ebf_getcheck_units(tagnames,dataunit=dataunit)
           if keyword_set(Append) eq 0 then append=0

           for i=0,n_elements(tagnames)-1 do begin
              ind=where(strcmp(datatags,tagnames[i],/FOLD_CASE),count)
              if count eq 1 then begin
                 name=byte(aliases1[i]) 
                 if (size(mydata[datatags[ind[0]]],/type) eq 8)or (size(mydata[datatags[ind[0]]],/type) eq 11) then begin
                    ebf_write_multi,fname1,path+aliases1[i]+'/',mydata[datatags[ind[0]]],REORDER=reorder,APPEND=Append,DATAUNIT=dataunit1.(i),nochar=nochar
                 endif else begin                 
                    ebf_write_single,fname1,path+aliases1[i],mydata[datatags[ind[0]]],APPEND=append,REORDER=Reorder,DATAUNIT=dataunit1.(i),nochar=nochar
                endelse    
                 if append eq 0 then append=1
                 status=1
              endif
              if count gt 1 then message,'EBF error: multiple tags'
           endfor           
    endif else message,'data must be a structure or hash or hashtable'   

end




pro ebf_write,fname1,path1,mydata,Tagnames=Tagnames,Aliases=Aliases,REORDER=reorder,APPEND=Append,STATUS=Status,VERBOSE=Verbose,DATAUNIT=dataunit,VERSION=Version,nochar=nochar
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  Version='0.0.3'
  if ~keyword_set(Append) then begin
     openw,lun,fname1,error=err,/GET_LUN     
     if(err ne 0) then begin
        print,err
        if n_elements(lun) ne 0 then free_lun,lun
     endif else free_lun,lun   
     ebftable_init,fname1
     append=1
  endif   

  temp=byte(path1) 
  if keyword_set(Tagnames)or(temp[n_elements(temp)-1] eq 47) then begin
     ebf_write_multi,fname1,path1,mydata,REORDER=reorder,STATUS=Status,VERBOSE=Verbose,Tagnames=Tagnames,Aliases=Aliases,APPEND=Append,DATAUNIT=dataunit,nochar=nochar
  endif else begin
     ebf_write_single,fname1,path1,mydata,REORDER=reorder,STATUS=Status,VERBOSE=Verbose,APPEND=Append,DATAUNIT=dataunit,nochar=nochar
  endelse

end



