;----------------------------------------------------------------------------
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;* Author: Sanjib Sharma                                                     *
;* Copyright (c) 2012 Sanjib Sharma                                          *
;* All rights reserved.                                                      *
;*                                                                           *
;* This file is part of EBF.  The full EBF copyright notice, including       *
;* terms governing use, modification, and redistribution, is contained in    *
;* the files COPYING and COPYRIGHT, which can be found at the root           *
;* of the source code distribution tree.                                     *
;* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
;-----------------------------------------------------------------------------
;+
;
; NAME:
;       EBFTABLE_PUT
;
; PURPOSE: 
;       Put the data name and its location into the hashtable.
;       
; CALLING SEQUENCE:
;       ebftable_put(filename,key,value,ecode)
;
; PARAMS:
;       fname : A string specifying the filename to write to.
;       key   : A string specifying the data name.
;       value : An integer specifying the location in file.
;       ecode : Is an output variable giving Error code. If there 
;               is no error then it is 0
;
; EXTERNAL PROGRAMS USED:
;       ebftable_setup.pro
;
; MODIFICATION HISTORY:
;       Written, 2013, Sanjib Sharma
;-

function ebftable_check_capacity,header,keys
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  if (header.current+n_elements(keys)+1) gt header.itemcapacity then return,0
  if (header.keyposcur+strlen(strjoin(keys))+1) gt header.keycapacity then return,0
  return,1
end

function ebftable_ebfckhash,mystr,hash1
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  if hash1 eq 0 then ehash=5381LL else ehash=long64(hash1)
  y=long64(byte(mystr))
  for i=0ULL,n_elements(y)-1ULL do begin
     ehash=ehash*33LL+y[i]
  endfor
  return,long64(ehash)
end


pro ebftable_expand,lun,dinfo,dinfo_pos,dflagswap,header,hflagswap,inkeys,fname
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  ebftable_getkeyvalsHT,lun,dinfo,header,hflagswap,keys,values
  ;; get name for trashing
  loc=0L
  i=-1L
  while loc ge 0 do begin
     i=i+1LL
     ebftable_get_keyval,lun,dinfo,header,hflagswap,'/.tr/.ebf/htable.t'+strcompress(string(i),/remove_all),loc
  endwhile


  ;; rename previous htable
  point_lun,lun,dinfo[1]
  ebfheader_read,ebfh,lun
  keyt='/.tr/.ebf/htable.t'+strcompress(string(i),/remove_all)
  ebfheader_rename,ebfh,keyt
  point_lun,lun,dinfo[1]
  ebfheader_write,ebfh,lun
  keys=[keys,keyt]
  values=[values,dinfo[1]]
  

  ;; get new capacity
  capacity=header.itemcapacity
  if capacity lt 0 then capacity=16LL
  while (n_elements(inkeys)+n_elements(keys)+1) gt capacity do capacity=capacity*2
  factor=header.keycapacity/header.itemcapacity    
  while (strlen(strjoin(inkeys))+strlen(strjoin(keys))+1) gt capacity*factor do capacity=capacity*2

  ebftable_create,lun,dinfo,dinfo_pos,dflagswap,header,hflagswap,capacity

  ind=where(keys eq '/.ebf/htable',count)
  if count gt 0 then values[ind[0]]=dinfo[1]

  for i=0LL,n_elements(keys)-1 do ebftable_add_keyval,lun,dinfo,header,hflagswap,keys[i],values[i]

end


pro ebftable_add_keyval,lun,dinfo,header,hflagswap,key,value
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  for i=0LL,n_elements(key)-1 do begin
     header.current=header.current+1
; check for space
; +1 needed as index is 1 based 
     if (header.current+1) gt header.itemcapacity then message,"not enough space for more keys"
; +1 not needed as index is zero based, but is just kept for simplicity"""
     if (header.keyposcur+strlen(key[i])+1) gt header.keycapacity then message,"not enough space for more keys"

; create item
     item={keyloc:0LL,keysize:0LL,value:0LL,next:0LL,tnext:0LL}
     item.keyloc=header.keyposcur
     item.keysize=strlen(key[i])
     item.value=value[i]
     item.next=-1
     item.tnext=-1
     header.keyposcur=header.keyposcur+strlen(key)          

; write current number of items
     point_lun,lun,dinfo[2]
     if hflagswap eq 1 then begin
        writeu,lun,swap_endian(header)
     endif else begin
        writeu,lun,header
     endelse            
     ebftable_write_node,lun,dinfo,header,hflagswap,header.current,item
     ebftable_write_key,lun,dinfo,header,hflagswap,item,key

; write to hash table or item updated with pointer
     keyhash=ebftable_ebflthash(key,header.htcapacity)
     loc1=ebftable_read_hvalue(lun,dinfo,header,hflagswap,keyhash)
     if loc1 ne 0 then begin
        item1=ebftable_read_node(lun,dinfo,header,hflagswap,loc1)
        while item1.next ne -1 do begin
           loc1=item1.next
           item1=ebftable_read_node(lun,dinfo,header,hflagswap,loc1)
        endwhile
        item1.next=header.current
        ebftable_write_node,lun,dinfo,header,hflagswap,loc1,item1
     endif else begin
        ebftable_write_hvalue,lun,dinfo,header,hflagswap,keyhash,header.current
     endelse
  endfor
end


pro ebftable_put,fname,keys1,values,ecode
  COMPILE_OPT STRICTARR,STRICTARRSUBS
  ON_ERROR,2
  ecode=0
  keys=strlowcase(keys1)
  ebftable_setup,fname,lun,dinfo,dinfo_pos,dflagswap,header,hflagswap,ecode

  if ecode eq 0 then begin
; check if key present
;     loc=ebftable_get_keyval(lun,dinfo,header,hflagswap,keys)
     ebftable_get_keyval,lun,dinfo,header,hflagswap,keys,loc

     if max(loc) ge 0 then begin
        print,'EBF->ebftable_put() Warning: one or more key already present'
     endif else begin
; check capacity
        if ebftable_check_capacity(header,keys) ne 1 then ebftable_expand,lun,dinfo,dinfo_pos,dflagswap,header,hflagswap,keys,fname

        for i=0LL,n_elements(keys)-1 do begin
           ebftable_add_keyval,lun,dinfo,header,hflagswap,keys[i],values[i]
        endfor
     endelse
  endif

  if ecode lt 100 then begin
     cksum=dinfo[0]
; compute checksum
     for i=0LL,n_elements(keys)-1 do begin
        mystr='('+keys[i]+', '+strcompress(string(values[i]),/remove_all)+')'
        cksum=ebftable_ebfckhash(mystr,cksum)
     endfor

; update checksum
     dinfo[0]=cksum
     point_lun,lun,dinfo_pos
     if dflagswap eq 1 then writeu,lun,swap_endian(dinfo[0]) else writeu,lun,dinfo[0]
  endif
  
  if n_elements(lun) ne 0 then free_lun,lun

  if ecode ge 1000 then begin 
     message,'Ebf error: cannot open file'
  endif
end
